import logging

from google.cloud import bigquery
from google.cloud.exceptions import GoogleCloudError

# Create logger object
logger = logging.getLogger(__name__)


def load_to_bigquery(
        client: bigquery.Client,
        dataset: str,
        table: str,
        uri: str,
        partition_field: str,
        clustering_fields: list = None,
        schema: list = None) -> None:
    """
    Loads a Cloud Storage object into a BigQuery partition.

    This function allows for highly consistent data loads
    into a ``DATE`` partition without the need of running
    an additional deduplication query.

    However, in order for the data load to be consistent,
    the partition date must representatively match the
    data it loads. Deviations from this will cause data
    to be placed in wrong partitions and result in,
    potentially, erroneous queries.

    Parameters
    ----------
    client : google.cloud.bigquery.client.Client
        The Google BigQuery client object.
    dataset : str
        The name of the destination dataset.
    table : str
        The name of the destination table.
    uri : str
        The URI or glob pattern of Cloud Storage object.
    partition_field : str
        The name of the column to be used as the
        parition_field.
    clustering_fields : list, optional
        The list of column names, in sequence, to be
        used a clustering fields.
    schema : list, optional
        The list of schema fields, in sequence, to be
        used when loading into the BigQuery table.
    """

    # Set job configuration
    logger.info("Generating LoadJobConfig")
    job_config = bigquery.LoadJobConfig()
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE
    job_config.time_partitioning = bigquery.TimePartitioning(
        field=partition_field,
        # TODO: PendingDeprecationWarning: Switch to Table.require_partition_filter
        require_partition_filter=True,
    )
    job_config.clustering_fields = clustering_fields
    job_config.schema = schema
    job_config.skip_leading_rows = 1
    logger.debug(f"LoadJobConfig properties: {vars(job_config)}")

    # Load to BigQuery
    logger.info("Starting BigQuery Job.")
    dataset_ref = client.dataset(dataset)
    table_ref = dataset_ref.table(table)
    job = client.load_table_from_uri(
        source_uris=uri,
        destination=table_ref,
        job_config=job_config,
    )

    # Waits for job to finish
    try:
        logger.info("Waiting for BigQuery Job to complete.")
        job.result()
    except GoogleCloudError:
        logger.critical("Encountered BigQuery loading errors!")
        for error in job.errors:
            logger.critical(error)
        raise

    # Update job status
    logger.info(f"Current BigQuery Job status: {job.state}")
    if job.state == "DONE":
        logger.info(f"Successfully uploaded '{uri}' to '{dataset}.{table}' on Google BigQuery.")
