import logging
import os

from google.cloud import storage
from google.cloud.exceptions import NotFound

# Create logger object
logger = logging.getLogger(__name__)


def download_from_storage(
        client: storage.Client,
        bucket_name: str,
        blob_name: str,
        destination_filename: str) -> None:
    """
    Helper function to download file from Google Cloud Storage.

    Parameters
    ----------
    client : google.cloud.storage.client.Client
        The Google Cloud Storage client object.
    bucket_name : str
        The name of the Google Cloud Storage source bucket.
    blob_name : str
        The name of the Google Cloud Storage source blob object.
    destination_filename : str
        The name of the destination file when downloaded
        into the local file system.

    Raises
    ------
    google.cloud.exceptions.NotFound
        If either bucket or blob does not exist.
    """

    # Get blob object
    logger.debug(f"Getting bucket and blob object.")
    bucket = client.get_bucket(bucket_name)
    blob = bucket.get_blob(blob_name)
    if not blob:
        raise NotFound(f"The blob '{blob_name}' does not exist.")

    # Downloading file
    # TODO: Verify hash?
    logger.info(f"Attempting to download 'gs://{bucket_name}/{blob_name}' from Google Cloud Storage into '{os.getcwd()}/{destination_filename}'.")
    blob.download_to_filename(destination_filename)

    # Check if file was indeed downloaded
    if not os.path.isfile(destination_filename):
        raise FileNotFoundError(f"File was supposedly downloaded to {destination_filename} but was not found!")

    # Log successful download
    logger.info(f"Successfully downloaded 'gs://{bucket_name}/{blob_name}' from Google Cloud Storage into '{os.getcwd()}/{destination_filename}'.")


def upload_to_storage(
        client: storage.Client,
        source_filename: str,
        bucket_name: str,
        blob_name: str) -> None:
    """
    Helper function to upload file to Google Cloud Storage.

    Parameters
    ----------
    client : google.cloud.storage.client.Client
        The Google Cloud Storage client object.
    source_filename : str
        The name of the source file in the local file system
        to be uploaded to Google Cloud Storage.
    bucket_name : str
        The name of the Google Cloud Storage destination bucket.
    blob_name : str
        The name of the Google Cloud Storage destination blob object.
    """

    # # Check if source file already exists
    # if not os.path.isfile(source_filename):
    #     raise FileNotFoundError(f"Attempted to upload from {source_filename} but it was not found!")

    # # Supply crc32c hash before upload
    # # NOTE: Supplying the hash allows GCS to verify the checksum and delete
    # # an uploaded file object if the checksum do not match. Hence, there is no
    # # need to validate the checksum locally for uploads.
    # logger.info("Assigning crc32c hash to blob")
    # blob.crc32c = generate_crc32_hash(
    #     filename=filename,
    #     algorithm="crc-32c",  # GCS uses CRC32C
    # )
    # logger.debug(f"CRC32c Hash: {hash}")

    # Get blob object
    bucket = client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    # # Warn if blob already exists
    # # NOTE: This logic assumes that it is ok to overwrite an existing object
    # # or that we already have object versioning enabled in Cloud Storage.
    # if blob.exists():
    #     logger.warn(f"Attempting to upload to {bucket_name}.{blob_name} but the blob object already exists.")

    # Upload file to Google Cloud Storage
    logger.info(f"Attempting to upload '{os.getcwd()}/{source_filename}' to 'gs://{bucket_name}/{blob_name} on Google Cloud Storage'.")
    blob.upload_from_filename(source_filename)

    # # Check if blob object exists
    # blob = bucket.get_blob(blob_name)
    # if not blob:
    #     raise NotFound(f"File was supposedly uploaded to {bucket_name}.{blob_name} but it does not exist.")

    # Log successful upload
    logger.info(f"Successfully uploaded '{os.getcwd()}/{source_filename}' to 'gs://{bucket_name}/{blob_name} on Google Cloud Storage'.")
