import datetime
import logging
import uuid

import pytz
from google.cloud import bigquery
from google.cloud import storage

from konekuta.utils import (
    get_reference_date,
    get_relative_date_range,
)
from konekuta.transformers import (
    split_data,
)
from konekuta.helpers.storage import (
    upload_to_storage,
)
from konekuta.helpers.bigquery import (
    load_to_bigquery,
)

# Create logger object
logger = logging.getLogger(__name__)


class Pipeline:
    """A base class for constructing ELT data pipelines."""

    def __init__(
            self,
            gcs_bucket: str,
            gcs_directory: str,
            bq_dataset: str,
            bq_table: str,
            bq_schema: list = None,
            bq_partitioning_field: str = None,
            bq_clustering_fields: list = None,
            gcloud_labels: dict = None,
            start_date: str = None,
            end_date: str = None,
            timezone: str = None,
            lookback: int = 1) -> None:
        """
        Create an instance of Pipeline.

        Parameters
        ----------
        gcs_bucket : str
            The name of the Google Cloud Storage bucket.
        gcs_directory : str
            The name of the Google Cloud Storage directory prefix.
            This name will be used as a prefix for all files uploaded.
        bq_dataset : str
            The name of the Google BigQuery dataset.
        bq_table : str
            The name of the Google BigQuery table.
        bq_schema : list, optional
            The schema of the source file and destination file.
        bq_partitioning_field : str, optional
            The partitioning field to be used when uploading the data
            into Google BigQuery.
        bq_clustering_fields : list, optional
            The list of clustering fields, in order, to be used when
            uploading the data into Google BigQuery.
        gcloud_labels : dict, optional
            The labels to be included in all Google Cloud resources
            interactions where possible.
        start_date : str, optional
            The start date (YYYY-MM-DD) used when querying data from
            a data source.
        end_date : str, optional
            The end date (YYYY-MM-DD) used when querying data from
            a data source.
        timezone : str, optional
            The timezone used for deriving the reference date when start
            and end date ranges are not supplied or are invalid.
        lookback : int, optional
            The lookback window, in days, used for deriving the start and
            end date ranges when they are not supplied or are invalid.
        """

        # Assign variables as instance attributes
        self.gcs_bucket = gcs_bucket
        self.gcs_directory = gcs_directory
        self.bq_dataset = bq_dataset
        self.bq_table = bq_table
        self.bq_schema = bq_schema
        self.bq_partitioning_field = bq_partitioning_field
        self.bq_clustering_fields = bq_clustering_fields
        self.gcloud_labels = gcloud_labels
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.lookback = lookback

        # Instantiate filenames
        self.raw_filename = None
        self.working_filename = None
        self.transformed_filenames = None

        # Set instance identifiers
        self.instance_timestamp = datetime.datetime.now(pytz.utc)
        logger.debug(f"Instance Timestamp: {self.instance_timestamp}")
        self.instance_id = uuid.uuid4()
        logger.debug(f"Instance ID: {self.instance_id}")

        # Validate and create tzinfo object
        # TODO: Add tests
        logger.debug("Parsing timezone parameter.")
        if self.timezone not in pytz.all_timezones:
            logger.warning(f"Timezone '{self.timezone}' was not found in pytz.all_timezones, defaulting to 'UTC'.")
            self.timezone = "UTC"
        self.timezone = pytz.timezone(self.timezone)
        logger.debug(f"Timezone: {self.timezone}")

        # Set reference date
        # TODO: Add tests
        logger.debug("Getting reference date from current timestamp and timezone.")
        self.reference_date = get_reference_date(
            timestamp=self.instance_timestamp,
            timezone=self.timezone,
        )
        logger.debug(f"Localised Reference Date: {self.reference_date}")

        # Override start_date and end_date if either are not specified
        # TODO: Handle if start_date and end_date are not datetime objects
        # TODO: Add tests
        logger.debug("Checking if both 'start_date' and 'end_date' variables are provided.")
        if not self.start_date or not self.end_date:
            logger.info(f"Either 'start_date' or 'end_date' was not provided!")

            # Get date ranges
            logger.info(f"Deriving date ranges based on lookback period of {self.lookback} day(s).")
            self.start_date, self.end_date = get_relative_date_range(
                date=self.reference_date,
                lookback=self.lookback
            )
            logger.debug(f"Setting date ranges: {self.start_date} -> {self.end_date}")

        # The following uses credentials inferred from the local environment
        # using Application Default Credentials.
        # https://googleapis.dev/python/google-api-core/latest/auth.html
        self.bigquery_client = bigquery.Client()
        self.storage_client = storage.Client()

    def run(self):
        """Executes the full ETL data pipeline."""

        self.extract()
        self.transform()
        self.load()

    def extract(self):
        """
        Extracts raw data from a data source.

        By default, the ``extract()`` method expects the
        ``self.raw_filename`` attribute to be assigned. The
        file bearing value of this attribute will then be
        uploaded to Cloud Storage in its raw format.

        Note
        -----
        Currently, while the method does not require the raw
        file to be in any format, it only supports **a single
        file** at this point in time.
        """

        # Handle if self.raw_filename is None
        logger.debug("Checking if instance attribute 'raw_filename' was assigned.")
        if not self.raw_filename:
            # TODO: Custom exception
            raise Exception("Instance attribute 'raw_filename' was not assigned!")

        # Upload raw file to Google Cloud
        upload_to_storage(
            client=self.storage_client,
            source_filename=self.raw_filename,
            bucket_name=self.gcs_bucket,
            blob_name=f"{self.gcs_directory}/raw/{self.instance_timestamp.astimezone(pytz.utc).isoformat()}",
        )

    def transform(self):
        """
        Transforms raw data into desired ingestion format.

        By default, the ``transform()`` method splits a CSV
        file into multiple date-level CSV files, before uploading
        all the files to Cloud Storage.

        Note
        -----
        Currently, the method only accepts a **single, uncompressed,
        UTF8-encoded CSV file** with **no leading or trailing
        metadata rows**. Additionally, all dates must be **in
        ``YYYY-MM-DD`` format**.

        References
        ----------
        https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        """

        # Handle if self.working_filename is None
        logger.debug("Checking if instance attribute 'working_filename' was assigned.")
        if not self.working_filename:
            logger.warning(f"Instance attribute 'working_filename' was not assigned, falling back to '{self.raw_filename}'.")
            self.working_filename = self.raw_filename

        # Split data file into multiple rows
        logger.debug(f"Splitting '{self.working_filename}' into multiple date-partitioned CSV files using the '{self.bq_partitioning_field}' column.")
        self.transformed_filenames = split_data(
            filename=self.working_filename,
            fieldname=self.bq_partitioning_field,
        )

        # Upload all partitioned files to GCS
        logger.debug(f"Uploading files in queue: {self.transformed_filenames}")
        for file_ in self.transformed_filenames:
            upload_to_storage(
                client=self.storage_client,
                source_filename=file_,
                bucket_name=self.gcs_bucket,
                blob_name=f"{self.gcs_directory}/{file_}",
            )

    def load(self):
        """
        Loads transformed data into BigQuery.

        By default, the ``load()`` method executes a LoadJob that
        imports all CSV files within a Cloud Storage directory
        into a designated BigQuery table.
        """

        # Load all files to BQ
        uri_pattern = f"gs://{self.gcs_bucket}/{self.gcs_directory}/*.csv"
        load_to_bigquery(
            client=self.bigquery_client,
            dataset=self.bq_dataset,
            table=self.bq_table,
            uri=uri_pattern,
            partition_field=self.bq_partitioning_field,
            schema=self.bq_schema,
        )
