import csv
import os
import datetime
import logging

# Create logger object
logger = logging.getLogger(__name__)


def split_data(
        filename: str,
        fieldname: str = "Date",
        pattern: str = "%Y-%m-%d") -> list:
    """
    Splits a date-level level dataset into multiple files.

    This function is meant to support a date-partitioned
    date warehouse design. Regardless of the date pattern
    in the input file, the name of the output file will
    always follow the ISO 8601 date format.

    Parameters
    ----------
    filename : str
        The name of the source file.
    fieldname : str
        The name of column header containing date strings.
        Defaults to ``Date``.
    pattern : str, optional
        The pattern used to interpret the string-formatted
        dates. Defaults to ``"%Y-%m-%d"``.

    Returns
    -------
    list
        The list of names of files created by the function.
    """

    # Loop every row in file
    files = set()
    with open(filename, "r") as fi:
        reader = csv.DictReader(fi)
        for row in reader:

            # Open corresponding file
            # TODO: Consider using datetime.date.fromisoformat()
            # when support for Python 3.6 is dropped.
            d = datetime.datetime.strptime(row[fieldname], pattern).date()
            df = f"{d.isoformat()}.csv"
            mode = "x"

            # Check if file exists before
            created = df in files
            exists = os.path.isfile(df)
            if exists and created:
                mode = "a"

            # Open file
            with open(df, mode=mode) as fo:
                writer = csv.DictWriter(fo, fieldnames=reader.fieldnames)

                # Write header if file does not exists before
                if not exists:
                    writer.writeheader()
                    files.add(df)

                # Write row data
                writer.writerow(row)

    # Return set as list
    return list(files)
