import datetime
import logging

# Create logger object
logger = logging.getLogger(__name__)


def get_reference_date(timestamp: datetime.datetime, timezone: datetime.tzinfo) -> datetime.date:
    """
    Returns a localised reference date based on a supplied
    timezone.

    The ``timestamp`` variable supplied to the method should be
    timezone aware.

    Parameters
    ----------
    timestamp : datetime.datetime
        The timezone-aware timestamp to be used to derive the
        reference date.
    timezone : datetime.tzinfo
        The timezone used to localise the timestamp.

    Returns
    -------
    datetime.date
        Returns the date of a localised timestamp.
    """

    # Type checking
    if not isinstance(timestamp, datetime.datetime):
        raise TypeError(f"Expected argument 'timestamp' to be an instance of datetime.datetime object: {timestamp} ({type(timestamp)})")
    if not isinstance(timezone, datetime.tzinfo):
        raise TypeError(f"Expected argument 'timezone' to be an instance of datetime.tzinfo object: {timezone} ({type(timezone)})")

    # Check if timestamp is timezone aware
    # TODO: Custom exception
    if not timestamp.tzinfo:
        raise Exception(f"Expected argument 'timestamp' to be timezone-aware")

    # Localise timestamp
    logger.info(f"Localising timestamp {timestamp} to timezone {timezone}")
    localised = timestamp.astimezone(timezone)
    logger.debug(f"Localised Timestamp: {localised}")

    # Return localised timestamp as date
    return localised.date()


def get_relative_date_range(date, lookback=1):
    """
    Returns a relative date range based on a reference date and
    the lookback period.

    The static method only deals with date objects, which means
    that the date supplied as an argument has to be already resolved
    in the desired timezone.

    Parameters
    ----------
    date : datetime.date
        The date to be used a reference point.
    lookback : int, optional
        The desired lookback period for date ranges.

    Returns
    -------
    tuple
        Returns a tuple of start date and end date as datetime.date
        objects.
    """

    # Type checking
    if not isinstance(date, datetime.date):
        raise TypeError(f"Expected argument 'date' to be an instance of datetime.date object: {date} ({type(date)})")
    if not isinstance(lookback, int):
        raise TypeError(f"Expected argument 'lookback' to be an instance of int object: {lookback} ({type(lookback)})")

    # Set lookback to default if out of range
    if lookback < 1:
        default_lookback = 1
        logger.warning(f"The specified lookback period of {lookback} day(s) is invalid, defaulting to lookback period of {default_lookback}")
        lookback = default_lookback

    # Generate date ranges
    start_date = date - datetime.timedelta(days=lookback)
    end_date = date - datetime.timedelta(days=1)

    # Return tuple of date ranges
    return start_date, end_date
